

/*******************************************************************/
/* 2-D parabolic equation                                          */
/*    implicit scheme                                              */
/*                                                                 */
/* Use                                                             */
/*                                                                 */
/* gcc 2d_diffusion_pipe.c -lm -o 2d_diffusion_pipe                */ 
/*                                                                 */
/* ./2d_diffusion_pipe N_x N_y | ./display_frames_GTK.py -s Nx Ny  */ 
/*                                                                 */ 
/*  ( suggested ( N_x ~ 2 N_y )                                    */
/*  ( Max dimension for Nx: 700 )                                  */
/*  ( N_x=700, N_y=400 is suggested )                              */
/*                                                                 */
/* display_frames_GTK.py was modified to read the field            */
/* in reverse order                                                */
/* and made exxecutable    chmod +x display_frames_GTK.py          */
/*                                                                 */
/* Compilation assumes subdirectory source                         */
/* containing 2d_source.c                                          */
/*                                                                 */
/*******************************************************************/

                                           /* Michel Vallieres      */
                                           /* modifiied Fall 2012   */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "../source/2d_source.c"


#define DEBUG 2                /* control extra printing */
#define IMAGE 1                /* control plot solution  */
#define SOURCE 1               /* control source term ( = 1 for source ) */

#define MAX_x 700              /* maximum dimensions   */
#define MAX_y 700
#define MAX_time_steps 3500    /* # time steps         */
#define MAX_iter 5000          /* maximum # iterations */
#define SMALL 0.000000001      /* required accuracy    */

#define alpha 0.03             /* time dependence of source */
#define t0 200                 /* for time domain [0,3000]  */
#define t1 500

int     sizex, sizey;          /* actual domain size - see    */
                               /* setup_domain_bc routine     */
double  rr;                    /* delta_t/2*h*h               */
double  delta_h;               /* h                           */
double  delta_t;               /* dt                          */

                               /* arrays for fields & source  */
                               /* and scratch use             */
double u[MAX_y][MAX_x], d[MAX_y][MAX_x];
double drexel[MAX_y][MAX_x];


                                     /* setup domain and    */
                                     /* boundary conditions */
void setup_domain_bc( int argc, char *argv[] )
{
   int    ix, iy;
   double field_top, field_bot, delta_field;

                                     /* domain parameters */
    if ( argc != 3 ) 
      {   
        printf( 
	"\n Syntax: 2d_parabolic  N_x  N_y\n  ( where N_x ~ 2 N_y ) \n\n" );
        exit(1);
      }
   sizex = atoi(argv[1]);
   sizey = atoi(argv[2]);
   fprintf( stderr, "\n Grid size: %d %d \n", sizex, sizey );
   if ( sizex > MAX_x || sizey > MAX_y )
     {
       fprintf( stderr, "Max lattice size: %d x %d\n\n", MAX_x, MAX_y);
       exit(1);
     }
   delta_h = 0.01;
   delta_t = 0.0002;
   rr = delta_t / ( 2 * delta_h * delta_h );
   fprintf( stderr, 
       "\n dt, h, r : %f %f %f \n\n", delta_t, delta_h, rr );

                                     /* boundary conditions */
   field_bot = 0.0;
   field_top = 1.0;
   for ( ix=0 ; ix<sizex ; ix++ )
     {
       u[0][ix] = field_top; 
       u[sizey-1][ix] = field_bot;
     }
   delta_field =  ( field_top - field_bot ) / ( sizey - 1 );
   for ( iy=1 ; iy<sizey-1 ; iy++ )
     {
       u[iy][0] = field_top - iy * delta_field;
       u[iy][sizex-1] = field_top - iy * delta_field;
     }
}


                                     /* initial fields */
void init_field()
{
   int    ix, iy;

   for ( iy=1 ; iy<sizey-1 ; iy++ )
     {
      for ( ix=1 ; ix<sizex-1 ; ix++ )
          u[iy][ix] = 0.0;
     }
}


                                     /* time dependence of source */
double time_source( int it )
{
   double value, t;
   t = it;
   value = 1.0/(1.0+exp(-alpha*(t-t0))) -
               1.0/(1.0+exp(-alpha*(t-t1)));
   return value;
}


                                     /* perform one time step */
int time_step( int it )
{
   int    ix, iy, iter;
   double temp, previous, max_change;
   double time_average, rhs_finite_diff, source_value;


                                     /* time average of      */
                                     /* source               */
   time_average = 0.0;
   if ( SOURCE == 1 ) 
       time_average = 0.5 * 
                  ( time_source(it-1) + time_source(it) );

                                     /* RHS at previous time */
                                     /* and source           */
   temp = 1 - 4*rr;
   for ( iy=1 ; iy<sizey-1 ; iy++ )
     {
       for ( ix=1 ; ix<sizex-1 ; ix++ )
         {
           rhs_finite_diff = temp * u[iy][ix] + 
                  rr * ( u[iy][ix+1] + u[iy][ix-1]  +
                              u[iy+1][ix] + u[iy-1][ix] );
           source_value =  drexel[iy][ix] * time_average;
           d[iy][ix] = rhs_finite_diff + 
                                delta_t * source_value;
         }
     }
                                     /* iterate implicit solution */
   temp = 1.0 / ( 1.0 + 4*rr );
   for ( iter=1 ;  iter <= MAX_iter ; iter++ )
     {
        max_change = 0.0;
	                             /* first lattice sweep */
        for ( iy=1 ; iy<sizey-1 ; iy++ )
          {
            for ( ix=1 ; ix<sizex-1 ; ix++ )
              {
                 previous = u[iy][ix];
                 u[iy][ix] = temp * (  d[iy][ix] + 
                      rr * ( u[iy][ix+1] + u[iy][ix-1]  +
                                 u[iy+1][ix] + u[iy-1][ix] )  );
                 if ( fabs( previous - u[iy][ix] ) > max_change )
                              max_change = fabs( previous - u[iy][ix] );
	      }
	  }
	                             /* second lattice sweep */
        for ( ix=sizex-2 ; ix>0 ; ix-- )
          {
            for ( iy=sizey-2 ; iy>0 ; iy-- )
              {
                 previous = u[iy][ix];
                 u[iy][ix] = temp * (  d[iy][ix] +
                      rr * ( u[iy][ix+1] + u[iy][ix-1]  +
                                 u[iy+1][ix] + u[iy-1][ix] )  );
                 if ( fabs( previous - u[iy][ix] ) > max_change )
                              max_change = fabs( previous - u[iy][ix] );
              }
          }
	                             /* converged ? */
          if ( max_change < SMALL )
            { 
              if ( DEBUG > 1 ) 
                fprintf(  stderr,  
                     "Time: %d -- iter: %d -- change: %g \n", 
                                        it, iter, max_change );
              return 0;
	    }       
     }
                                     /* no convergence */
     fprintf( stderr, " Failed to converge \n" );
     exit (1);
}


                                     /* record the field in individual */
                                     /* VTK files to make frames for   */
                                     /* a movie                        */
int display_field( char *fname )
{
   int    ix, iy, iy_inverted;
   int    nc;
   double vmin, vmax;
   float  nodal[ MAX_x * MAX_y ];    /* linear array -> pointer access */
 
                                     /* min & max and         */
                                     /* create nodal variable */
   vmin = u[0][0];
   vmax = u[0][0];
   for ( iy=0 ; iy<sizey ; iy++ )
     {
      for ( ix=0 ; ix<sizex ; ix++ )
        {
           if ( u[iy][ix] < vmin ) vmin =  u[iy][ix];
           if ( u[iy][ix] > vmax ) vmax =  u[iy][ix];
	   iy_inverted = sizey - 1 - iy;
	   nodal[ iy_inverted*sizex + ix ] = u[iy][ix];
        }
     }

  fprintf( stderr, " min & max %f %f -- file name: %s\n", 
                            vmin, vmax, fname );

   for ( iy=0 ; iy<sizey ; iy++ )
     {
      for ( ix=0 ; ix<sizex ; ix++ )
        {
	  printf(  "%f\n",  u[iy][ix] );
        }
     }
                                     /* write image data */

}



int main( int argc, char *argv[] )
{
   int  it;
   int  frame_count;
   char fname[256];

                                     /* domain and boundary cond */
   setup_domain_bc( argc, argv );

                                     /* set-up source */
   setup_source( SOURCE );
                                     /* initialize field */
   init_field();
                                     /* movie frame count */
   frame_count = 0;
                                     /* loop over time */
   for ( it=1 ; it<=MAX_time_steps ; it=it+1 )
     {
                                     /* perform 1 time step */
        time_step ( it );
                                     /* display field ever so often */
        if ( IMAGE > 0  && it % 10 == 0 ) 
	  {
                                    /* file name in # sequence */
             sprintf( fname, "./movie/frame%04d.vtk", frame_count );
	     frame_count++;
	                            /* write out data for frame */
             display_field( fname );
	  }
     }
 
   exit(0);
}

